/*****************************************************************************/
/*
                              CGIplusTest.c

Script for testing the relative efficiencies of CGI and CGIplus.
Also for scripting benchmarks between the WASD, OSU and Apache environments!

This script can detect whether it is in a CGI or CGIplus environment and
adjust behaviour accordingly.  Call using the AB (ApcheBench) utility in both
environments to test throughput and response latency performance in both.
For WASD and OSU this script re-opens <stdout> in binary-mode, hence the full
HTTP response.  For Apache it leaves it in record-mode and uses a CGI response. 
Requires the CGI environment variables to be available as symbols.

September 2000 Note: Apache 1.3-12 seems intent on adding a <CR><LF> to each
record received, distorting the content-length and generally causing problems. 
This seems to lead to ApacheBench reporting errors, including length issues.

Can also be used as example code for working in both environments.
Demonstrates the alternative mechanism for accessing CGIplus variable values,
reading from <stdin> and parsing each line for required variables.  Also see
CGIplusSkeleton.c for the more spohisticated, general-purpose function.


VERSION HISTORY (update SoftwareID as well!)
---------------
23-SEP-2000  MGD  v1.2,0, minor mods for comparison with VMS Apache
05-MAR-1999  MGD  v1.1.0, minor mods for comparison with OSU
08-JUN-1997  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "CGIPLUSTEST AXP-1.2.0";
#else
   char SoftwareID [] = "CGIPLUSTEST VAX-1.2.0";
#endif

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

int  ApacheEnvironment,
     Debug,
     IsCgiPlus,
     OsuEnvironment;

char  *CgiPlusEofPtr,
      *CgiServerSoftwarePtr;

char  Utility [] = "CGIPLUSTEST",
      Line80Chars [] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz\
0123456789\
!@#$%^&*()_+-=<>,\n";

/* function prototypes */
AtOsuExit ();

/*****************************************************************************/
/*
*/

main ()

{
   register char  *cptr;

   int  RequestedCount;

   /*********/
   /* begin */
   /*********/

   if ((CgiServerSoftwarePtr = getenv ("WWW_SERVER_SOFTWARE")) == NULL)
      if ((CgiServerSoftwarePtr = getenv ("SERVER_SOFTWARE")) == NULL)
         CgiServerSoftwarePtr = "?";

   if (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL)
   {
      OsuEnvironment = 1;
      atexit (&AtOsuExit);
   }
   else
   {
      if (strstr (CgiServerSoftwarePtr, "Apache") != NULL)
      ApacheEnvironment = 1;
   }
   
   if (getenv ("CGIPLUSTEST$DBUG") != NULL) Debug = 1;

   if (Debug)
   {
      /* won't work for OSU! */
      fprintf (stdout, "Content-Type: text/plain\r\n\r\n");
      system ("show sym *");
   }
   else
   {
      /* used to make comparisons with OSU and it spits with output > 4096 */
      if (OsuEnvironment)
      {
         /* reopen so '\r' and '\n' are not filtered, maximum record size */
         if ((stdout =
             freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin", "mrs=4096"))
             == NULL)
            exit (vaxc$errno);

         fprintf (stdout, "<DNETCGI>");
         fflush (stdout);
      }
      else
      if (!ApacheEnvironment)
      {
         /* reopen output stream so that the '\r' and '\n' are not filtered */
         if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin")) == NULL)
            exit (vaxc$errno);
      }
   }

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);

   /* do once for standard CGI, multiple times for CGIplus */

   do {

      if (IsCgiPlus)
      {
         /* called as a CGIplus script */
         RequestedCount = CgiPlusCount ();
      }
      else
      {
         /* called as a standard CGI script */
         if ((cptr = getenv ("WWW_QUERY_STRING")) == NULL)
            if ((cptr = getenv ("QUERY_STRING")) == NULL)
               cptr = "";
         if (isdigit(*cptr))
            RequestedCount = atoi(cptr);
         else
            RequestedCount = -1;
      }

      fprintf (stdout,
"Content-Type: text/plain\r\n\
Content-Length: %d\r\n\
Script-Control: X-stream-mode\r\n\
Expires: Fri, 13 Jan 1978 14:00:00 GMT\r\n\
\r\n",
         RequestedCount * (sizeof(Line80Chars)-1));

      if (RequestedCount < 0)
      {
         fprintf (stdout,
"Usage: Supply a URL with a query string containing \
an integer representing the number of 80 character lines \
to be returned during the test.\n");
      }
      else
      {
         while (RequestedCount--)
            fputs (Line80Chars, stdout);
      }

      if (IsCgiPlus)
      {
         /* the CGIplus EOF must be an independant I/O record */
         fflush (stdout);
         fprintf (stdout, "%s", CgiPlusEofPtr);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (1);
}

/*****************************************************************************/
/*
Script executing in OSU environment exits.
*/

AtOsuExit ()

{
   if (Debug)
      fprintf (stdout, "</DNETTEXT>\n");
   else
   {
      fflush (stdout);
      fprintf (stdout, "</DNETCGI>");
      fflush (stdout);
   }
}

/*****************************************************************************/
/*
Use a simple loop to read lines from standard input.  Detect specific CGI
variables by checking the leading section of these lines for the required
variable name.  The value may then be extracted from the remaining section.
*/

int CgiPlusCount ()

{
#  define CGIPLUS_LINE_SIZE 1024

   register char  *cptr;

   int  Count;
   char  Line [CGIPLUS_LINE_SIZE];
   FILE  *SysCommand;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiPlusCount()\n");

   Count = -1;

   if ((SysCommand = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
      exit (vaxc$errno);

   for (;;)
   {
      if (fgets (Line, sizeof(Line), SysCommand) == NULL) exit (vaxc$errno);
      if (Debug) fprintf (stdout, "Line |%s|\n", Line);
      /* first empty line signals the end of CGIplus variables */
      if (Line[0] == '\n') break;
      /* remove the trailing newline */
      if ((cptr = strchr(Line, '\n')) != NULL) *cptr = '\0';

      if (!strncmp (Line, "WWW_QUERY_STRING=", 17))
         if (isdigit(Line[17])) Count = atoi(Line+17);
   }

   fclose (SysCommand);

   return (Count);
}

/*****************************************************************************/

