/*****************************************************************************/
/*
                                Service.h

Function prototypes for service configuration module.

*/
/*****************************************************************************/

#ifndef SERVICE_H_LOADED
#define SERVICE_H_LOADED 1

#ifndef ODSSTRUCT_H_LOADED
#   include "odsstruct.h"
#endif

#include "wasd.h"

/**************/
/* structures */
/**************/

struct ServiceStruct
{
   struct ServiceStruct  *NextPtr;

   boolean  ConnectService,
            GenericService,
            LogFileError,
            LogFileOpen,
            NoLog,
            LocalAuthRequired,
            ProxyAuthRequired,
            ProxyFileCacheEnabled,
            ProxyService,
            ProxyTrack,
            SetNoTrack,
            SetProxyTrack,
            TrackEnabled;

   int  ClientPort,
        ConnectCount,
        ErrorReportPathLength,
        ListenBacklog,
        LogFileNameLength,
        ProxyChainHostNameLength,
        ProxyChainPort,
        RequestScheme,
        ServerHostNameLength,
        ServerPort;

   unsigned long  ClientIpAddress,
                  ProxyChainIpAddress,
                  ServerHostNameHashValue,
                  ServerHostPortHashValue,
                  ServerIpAddress;

   unsigned long  QuadBytesRawRx [2],
                  QuadBytesRawTx [2];

   char  *SeSoLaCertPtr,
         *SeSoLaCipherListPtr,
         *SeSoLaKeyPtr,
         *RequestSchemeNamePtr;

   char  ClientHostName [128],
         ClientIpAddressString [32],
         BodyTag [256],
         ErrorReportPath [128],
         LogFileName [256],
         LogHostName [64],
         ProxyChainHostName [128],
         ProxyChainHostPort [128+16],
         ProxyChainIpAddressString [32],
         ProxyChainPortString [16],
         ServerHostName [128],
         ServerHostPort [128+16],
         ServerIpAddressString [32],
         ServerPortString [16],
         SSLcert [128],
         SSLcipherList [128],
         SSLkey [128],
         SuppliedIpAddressString [32];

   struct AnIOsb  ServerIOsb;

   struct dsc$descriptor  LogFileNameDsc;
   
   struct FAB  LogFileFab;
   struct RAB  LogFileRab;

   /* this will be cast using (SSL_CTX*) in SESOLA.C */
   void  *SeSoLaCtx;

   unsigned short  ClientChannel,
                   ServerChannel;

   struct sockaddr_in  ServerSocketName,
                       ClientSocketName;

   struct {
      unsigned int  Length;
      char  *Address;
   } ServerSocketNameItem;

   int  ClientSocketNameLength;
   struct {
      unsigned int  Length;
      char  *Address;
      unsigned int  *LengthPtr;
   } ClientSocketNameItem;
};

struct ServiceLoadStruct
{
   struct ServiceStruct  *ListHead,
                         *ListTail;
   struct ServiceStruct  **HashTable;

   int  ServiceCount;

   /* basically only used while reading configuration from file */
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   char  *ProblemReportPtr;
   char  DirectiveName [128],
         LoadFileName [ODS_MAX_FILE_NAME_LENGTH+1];
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   /* only used when reporting or changing configuration */
   void  *RequestPtr;
}

/***********************/
/* function prototypes */
/***********************/

#endif /* SERVICE_H_LOADED */

/*****************************************************************************/
